/*
 * Copyright 2007-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com) 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.unitestng;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.ArrayList;
import java.util.List;

public class FileComparison
{
    public enum ComparisonMode { ENTIRE, TAIL }

    private static final String endl = System.getProperty( "line.separator", "\n" );

    public static void compareFilesByLines( File ref, File file )
        throws FileSizesAreDifferentException, LinesAreDifferentException, IOException
    {
        compareFilesByLines( ref, file, ComparisonMode.ENTIRE, null );
    }

    public static void compareFilesByLines( File ref, File file, ComparisonMode mode )
    	throws FileSizesAreDifferentException, LinesAreDifferentException, IOException
    {
    	compareFilesByLines(ref, file, mode, null);
    }
    
    public static void compareFilesByLines(
    	File ref,
    	File file,
    	ComparisonMode mode,
    	StringMatcher matcher
    )
    	throws FileSizesAreDifferentException, LinesAreDifferentException, IOException
    {
        int start_line
            = (mode == ComparisonMode.TAIL)
            ? FileUtil.calculateNumberOfLines( file ) - FileUtil.calculateNumberOfLines( ref ) + 1
            : 1;
        LineNumberReader file_reader = null;
        LineNumberReader ref_reader = null;
        try {
        	file_reader = new LineNumberReader( new FileReader( file ) );
        	ref_reader = new LineNumberReader( new FileReader( ref ) );
        	while( file_reader.ready() ) {
        		if( !ref_reader.ready() ) {
        			throw new FileSizesAreDifferentException(
        					  FileSizesAreDifferentException.Kind.EXPECTED_FILE_IS_SHORTER
        					, ref
        					, file
        					, file_reader.getLineNumber()
        			);
        		}
        		String file_line = file_reader.readLine();
        		if( file_reader.getLineNumber() < start_line ) {
        			continue;
        		}
        		String ref_line = ref_reader.readLine();
        		
        		boolean equal
        			= matcher == null
        			? ref_line.equals(file_line)
        			: matcher.matches(ref_line, file_line);
        		if( !equal ) {
        			throw new LinesAreDifferentException(
        					  ref
        					, file
        					, file_reader.getLineNumber()
        					, ref_line
        					, file_line
        			);
        		}
        	}
	        if( ref_reader.ready() ) {
	            throw new FileSizesAreDifferentException( FileSizesAreDifferentException.Kind.EXPECTED_FILE_IS_LONGER
	                    , ref
	                    , file
	                    , file_reader.getLineNumber()
	                    );
	        }
        } finally {
        	if (file_reader != null)	file_reader.close();
        	if (ref_reader != null)		ref_reader.close();
        }
    }

    public static void compareFilesContent( File ref, File file )
        throws IOException, FilesOrDirectoriesAreDifferentException
    {
        compareFilesContent( ref, file, ComparisonMode.ENTIRE, null );
    }

    public static void compareFilesContent( File ref, File file, ComparisonMode mode )
    	throws IOException, FilesOrDirectoriesAreDifferentException
    {
    	compareFilesContent(ref, file, mode, null);
    }
    
    public static void compareFilesContent(
    	File ref,
    	File file,
    	ComparisonMode mode,
    	StringMatcher matcher
    )
        throws IOException, FilesOrDirectoriesAreDifferentException
    {
        if( !file.exists() )
        {
            throw new FilesOrDirectoriesAreDifferentException
                ( file.getCanonicalPath() + ": file or directory expected to exist" );
        }
        if( ref.isDirectory() )
        {
            if( !file.isDirectory() )
            {
                throw new FilesOrDirectoriesAreDifferentException
                    ( file.getCanonicalPath() + ": directory expected, regular file found" );
            }
            return;
        } else {
            if( file.isDirectory() )
            {
                throw new FilesOrDirectoriesAreDifferentException
                    ( file.getCanonicalPath() + ": regular file expected, directory found" );
            }
        }
        try
        {
            compareFilesByLines( ref, file, mode, matcher );
        } catch( FileSizesAreDifferentException e ) {
            switch( e.getKind() )
            {
            case EXPECTED_FILE_IS_LONGER:
                throw new FilesOrDirectoriesAreDifferentException
                    (   file.getCanonicalPath() + ": file is too short ("
                      + e.getLineNum() + " lines), expected file is longer"
                    , e
                    );
            case EXPECTED_FILE_IS_SHORTER:
                throw new FilesOrDirectoriesAreDifferentException
                    (   file.getCanonicalPath() + ": file is too long , expected file has "
                      + e.getLineNum() + " lines only"
                    , e
                    );
            default:
                throw new FilesOrDirectoriesAreDifferentException
                    ( "Unexpected FileSizesAreDifferent.Kind: " + e.getKind()
                    , e
                    );
            }
        } catch( LinesAreDifferentException e ) {
            throw new FilesOrDirectoriesAreDifferentException
                (   file.getCanonicalPath() + ":" + e.getLineNum() + ": '" + e.getFileLine()
                  + "', expected: '" + e.getRefLine() + "'"
                , e
                );
        }
    }

    public static void compareAllFilesInSubdirs( File out_dir, File ref_dir )
        throws IOException, FilesOrDirectoriesAreDifferentException
    {
        compareAllFilesInSubdirs( out_dir, ref_dir, ComparisonMode.ENTIRE, null );
    }

    public static void compareAllFilesInSubdirs( File out_dir, File ref_dir, ComparisonMode mode )
    	throws IOException, FilesOrDirectoriesAreDifferentException
    {
    	compareAllFilesInSubdirs(out_dir, ref_dir, mode, null);
    }
    
    public static void compareAllFilesInSubdirs(
    	File out_dir,
    	File ref_dir,
    	ComparisonMode mode,
    	StringMatcher matcher
    )
        throws IOException, FilesOrDirectoriesAreDifferentException
    {
        List<File> files_to_compare = new ArrayList<File>();
        FileUtil.collectFilesInSubdir( files_to_compare, ref_dir, new FileUtil.MaskFilenameFilter( ".+", "CVS|\\.svn" ) );
        int ref_dir_full_path_len = ref_dir.getCanonicalPath().length();
        StringBuffer sb = new StringBuffer();
        boolean is_different = false;
        for( File f : files_to_compare )
        {
            String file_name_tail = f.getCanonicalPath().substring( ref_dir_full_path_len );
            try
            {
                compareFilesContent( f, new File( out_dir.getCanonicalPath() + file_name_tail ), mode, matcher );
            } catch( FilesOrDirectoriesAreDifferentException e ) {
                if( is_different )
                {
                    sb.append( endl );
                }
                is_different = true;
                sb.append( e.getMessage() );
            }
        }
        if( is_different )
        {
            throw new FilesOrDirectoriesAreDifferentException( sb.toString() );
        }
    }

    //----  exceptions  ----

    public static class FilesOrDirectoriesAreDifferentException extends Exception
    {
        public FilesOrDirectoriesAreDifferentException( String msg )
        {
            super( msg );
        }
        public FilesOrDirectoriesAreDifferentException( String msg, Throwable cause )
        {
            super( msg, cause );
        }
    }

    public static abstract class FilesAreDifferentException extends Exception
    {
        private File ref;
        private File file;
        private int lineNum;

        public FilesAreDifferentException( File ref, File file, int line )
        {
            this.ref = ref;
            this.file = file;
            this.lineNum = line;
        }

        public File getRef() { return ref; }

        public File getFile() { return file; }

        public int getLineNum() { return lineNum; }
    }

    public static class FileSizesAreDifferentException extends FilesAreDifferentException
    {
        private static final long serialVersionUID = -8397840198572872055L;

        public enum Kind { EXPECTED_FILE_IS_LONGER
                         , EXPECTED_FILE_IS_SHORTER
                         , LINES_ARE_DIFFERENT
                         }

        private Kind kind;

        public FileSizesAreDifferentException( Kind kind, File ref, File file, int line )
        {
            super( ref, file, line );
            this.kind = kind;
        }

        public Kind getKind() { return kind; }
    }

    public static class LinesAreDifferentException extends FilesAreDifferentException
    {
        private static final long serialVersionUID = -4022825091738009649L;

        private String refLine;
        private String fileLine;

        public LinesAreDifferentException( File ref, File file, int lineNumber, String refLine, String fileLine )
        {
            super( ref, file, lineNumber );
            this.refLine = refLine;
            this.fileLine = fileLine;
        }

        public String getRefLine()
        {
            return refLine;
        }

        public String getFileLine()
        {
            return fileLine;
        }
    }
    
    public static interface StringMatcher {
    	boolean matches(String ref, String str);
    }
}
