/*
 * Copyright 2007-2009 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com) 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.unitestng;

import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.testng.Reporter;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

public abstract class UniTestNG extends TestSuiteBase
{
    protected File testGroupDir;
    protected Properties testGroupProperties;
    protected File testRootDir;
    protected String testDirMask;
    protected String nonTestDirMask;
    protected String testPropertiesFileName;
    
    public File getTestGroupDir() { return testGroupDir; }
    public File getTestRootDir() { return testRootDir; }
    
    /**
     *   ,   .
     */
    public void init( File testGroupDir, Properties testGroupProperties )
    {
        this.testGroupDir = testGroupDir;
        this.testGroupProperties = testGroupProperties;
        this.testRootDir
            = new File( this.testGroupDir, testGroupProperties.getProperty( "test.dir", "." ) );
        this.testDirMask = testGroupProperties.getProperty( "test.dir.mask", ".+" );
        this.nonTestDirMask = testGroupProperties.getProperty( "non.test.dir.mask", null );
        this.testPropertiesFileName
            = testGroupProperties.getProperty( "test.properties", "test.properties" );
    }
    
    /**
     *   ,     .
     */
    public void init( String fileName ) throws IOException
    {
        InputStream in = new FileInputStream( fileName );
        Properties p = new Properties(); 
        p.load( in );
        init( new File( fileName ).getParentFile(), p );
    }
    
    protected FileFilter testDirFilter 
        = new FileFilter() 
              { 
                  public boolean accept( File f ) 
                  { 
                      try
                      {
                          return    f.isDirectory() 
                                 && !f.isHidden() 
                                 && f.getCanonicalPath().matches( testDirMask )
                                 && (  nonTestDirMask == null
                                    || !f.getCanonicalPath().matches( nonTestDirMask )
                                    )
                                 ;
                      }
                      catch( IOException e )
                      {
                          e.printStackTrace();
                          return false;
                      } 
                  }
              };

    /**
     *        {@link #testRootDir}.
     *   .
     */
    @DataProvider( name = "testDirs" )
    public Object[][] getTestDirs() 
    {
        List<File> dirs = new ArrayList<File>();
        getTestDirs( dirs, testRootDir );
        int l = dirs.size();
        Object[][] res = new Object[l][];
        for( int i = 0; i < l; i++ )
        {
            res[i] = new Object[1];
            res[i][0] = dirs.get(i);
        }
        return res;
    }
    
    public void getTestDirs( List<File> dirs, File dir )
    {
        dirs.add( dir );
        for( File f : dir.listFiles( testDirFilter ) )
        {
            getTestDirs( dirs, f );
        }
    }
    
    /**
     *           ,   
     *   .
     * @throws IOException 
     * @throws FileNotFoundException 
     */
    void loadTestProperties( File dir, Properties testProperties )
        throws FileNotFoundException, IOException
    {
        if( !dir.getCanonicalFile().equals( testRootDir.getCanonicalFile() ) )
        {
            loadTestProperties( dir.getParentFile(), testProperties );
        }
        File testPropertiesFile = new File( dir, testPropertiesFileName );
        if( testPropertiesFile.exists() )
        {
            testProperties.load( new FileInputStream( testPropertiesFile ) );
        }
    }    

    /**
     *  .    ,   .
     *      {@link #testPropertiesFileName }
     *     .
     *      ,    ,
     *   {@link #init(File,Properties) }.
     * 
     * @throws ClassNotFoundException 
     * @throws IllegalAccessException 
     * @throws InstantiationException 
     */
    @Test( dataProvider="testDirs" )
    public void run( File dir ) throws Throwable
    {
        Properties testProperties = new Properties();
        loadTestProperties( dir, testProperties );
        
        String wrapper = testGroupProperties.getProperty( "run.wrapper" );
        if( wrapper != null ) runWrapper( testProperties, dir, wrapper );
//        String clazz = testDesc.getProperty( "run.class" );
//        if( clazz != null ) runClass( testDesc, dir, clazz );
//        String command = testDesc.getProperty( "run.command" );
//        if( command != null ) runCommand( testDesc, dir, command );
    }
    
    protected void runWrapper( Properties testProperties, File dir, String wrapperName )
        throws Throwable
    {
        Class<?> wrapperClass = Class.forName(  wrapperName );
        Object wrapper = wrapperClass.newInstance();
        try
        {
            ((Wrapper)wrapper).run( testProperties, dir );
        } 
        catch( ExitException e )
        {
            if( e.getStatus() != 0 ) throw e;
        }
    }
    
    public static class ExecResult
    {
        public int exitCode;
        public String stdout;
        public String stderr;
        
        public static final String STDOUT = "stdout"; 
        public static final String STDERR = "stderr"; 
        
        public void setStream( String name, String value )
        {
            if( STDOUT.equals( name ) ) stdout = value;
            else if( STDERR.equals( name ) ) stderr = value;
            else throw new IllegalArgumentException( name ); 
        }
    }

    public static int exec( String command, String[] env, File dir )
        throws IOException, InterruptedException
    {
        Reporter.log( "> " + command );
        Process p =  Runtime.getRuntime().exec( command, env, dir );
        logStream( p.getInputStream() );
        logStream( p.getErrorStream() );
        p.waitFor();
        return p.exitValue();
    }

    public static int exec( String[] cmdarray, String[] env, File dir )
        throws IOException, InterruptedException
    {
        return exec( cmdarray, env, dir, null );
    }

    public static int exec( String[] cmdarray, String[] env, File dir, ExecResult result )
        throws IOException, InterruptedException
    {
        StringBuilder sb = new StringBuilder();
        for( String s : cmdarray )
        {
            if( sb.length() > 0 ) sb.append( " " );
            sb.append( s );
        }
        Reporter.log( "> " + sb );
        Process p =  Runtime.getRuntime().exec( cmdarray, env, dir );
        Thread it = logStream( p.getInputStream(), ExecResult.STDOUT, result );
        Thread et = logStream( p.getErrorStream(), ExecResult.STDERR, result );
        p.waitFor();
        it.join();
        et.join();
        if( result != null ) result.exitCode = p.exitValue(); 
        return p.exitValue();
    }

    static class LogThread extends Thread
    {
        private InputStream in;
        private String stream;
        private ExecResult result;
        
        public LogThread( InputStream in )
        {
            this.in = in;
        }

        public LogThread( InputStream in, String stream, ExecResult result )
        {
            this( in );
            this.stream = stream;
            this.result = result;
        }

        @Override
        public void run()
        {
            StringWriter out = new StringWriter();
            int b;
            try
            {
                while( (b = in.read()) != -1 )
                {
                    out.write( b );
                }
            }
            catch( IOException e )
            {
                e.printStackTrace( new PrintWriter( out ) );
            }
            String res = out.getBuffer().toString();
            Reporter.log( res );
            if( result != null )
            {                
                result.setStream( stream, res );
            }
        }
    }
     
    public static Thread logStream( InputStream in )
    {
        Thread t = new LogThread( in );
        t.start();
        return t;
    }
    
    public static Thread logStream( InputStream in, String stream, ExecResult result )
    {
        Thread t = new LogThread( in, stream, result );
        t.start();
        return t;
    }
}
